#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <mpfi.h>
#include <mpfi_io.h>
#include <strtools.h>

// these extern values are assigned in example.c files.
extern const int DIMENSIONS;  // number of dimensions
extern const int ITERATIONS;  // number of iterations

extern const double BOX_K;    // size of initial box [-BOX_K, BOX_K]
extern const double MAX_TIME; // maximum time value

extern const int doLog;       // 0 means no log files; 1 means create log files
extern const int doPlot;      // 0 means no log files; 1 means create log files

extern char *equationName;    // mathematical fomula of the equation, printed on the output files
extern char *fileName;        // prefix for log and plt files

extern const int dim1;        // first dimension for 3d plot, usually 0
extern const int dim2;        // second dimension for 3d plot, usually 1
extern const double MIN_RES;  // small cubes are not visible in gnuplot

const double gnupltBoxCoeff = 1.05;
const mp_prec_t defaultPrecision = 64;

unsigned const int BASE = 10;
const size_t DIGITS = 5;

char **logFileName;
char **pltFileName;
char **datFileName;

// creates the output file names
void createFileNames(char * fileName) {
	logFileName = malloc(DIMENSIONS * sizeof (char *));
	pltFileName = malloc((DIMENSIONS+1) * sizeof (char *));
	datFileName = malloc((DIMENSIONS+1) * sizeof (char *));
	int ii;
	for (ii = 0; ii < DIMENSIONS; ii++) {
		char * st = malloc(20);
		itoa(ii, st, 10);

		if ((logFileName[ii] = malloc(30)) == NULL) {
			printf("Error.\n");
			exit(0);
		} else {
			strcpy(logFileName[ii], fileName);
			strcat(logFileName[ii], "_");
			strcat(logFileName[ii], st);
			strcat(logFileName[ii], ".log");
		}
		//printf("%s\n", logFileName[ii]);

		if ((pltFileName[ii] = malloc(30)) == NULL) {
			printf("Error.\n");
			exit(0);
		} else {
			strcpy(pltFileName[ii], fileName);
			strcat(pltFileName[ii], "_");
			strcat(pltFileName[ii], st);
			strcat(pltFileName[ii], ".plt");
		}
		//printf("%s\n", pltFileName[ii]);

		if ((datFileName[ii] = malloc(30)) == NULL) {
			printf("Error.\n");
			exit(0);
		} else {
			strcpy(datFileName[ii], fileName);
			strcat(datFileName[ii], "_");
			strcat(datFileName[ii], st);
			strcat(datFileName[ii], ".dat");
		}
		//printf("%s\n", datFileName[ii]);
	}

	// This creates the files for 3d plot
	if ((pltFileName[DIMENSIONS] = malloc(30)) == NULL) {
		printf("Error.\n");
		exit(0);
	} else {
		strcpy(pltFileName[DIMENSIONS], fileName);
		strcat(pltFileName[DIMENSIONS], ".plt");
	}

	if ((datFileName[DIMENSIONS] = malloc(30)) == NULL) {
		printf("Error.\n");
		exit(0);
	} else {
		strcpy(datFileName[DIMENSIONS], fileName);
		strcat(datFileName[DIMENSIONS], ".dat");
	}
}

// main
int main() {
	// creates the log and plt file names
	createFileNames(fileName);

	// some settings

	mp_rnd_t mp_rnd = GMP_RNDN; // will be used for mpfr rounding value

	mpfr_t max_time;
	mpfr_init(max_time);
	mpfr_set_d(max_time, MAX_TIME, mp_rnd);

	mpfr_set_default_prec(defaultPrecision);

	unsigned long k, kk;
	int i, j;

	// define currentVFvalue and nextVFvalue vectors
	mpfi_t currentVFvalue[DIMENSIONS], nextVFvalue[DIMENSIONS];

	// initialise the currentVFvalue and nextVFvalue vectors
	for (j = 0; j < DIMENSIONS; j++) {
		mpfi_init(currentVFvalue[j]);
		mpfi_init(nextVFvalue[j]);
	}

	mpfi_t box_K;
	mpfi_init(box_K);
	mpfi_interv_d(box_K, -BOX_K, BOX_K);

	mpfr_t temp_r;
	mpfr_init(temp_r);

	// initialise the currentVFvalue[i][j] for i \in {0, ..., INIT_PARTITIONS-1} to [-BOX_K, BOX_K]
	for (j = 0; j < DIMENSIONS; j++) {
		mpfi_set(currentVFvalue[j], box_K);
	}

	// define and initialise currentY and nextY
	mpfi_t currentY, nextY, derivative, integrate;
	mpfi_init(currentY);
	mpfi_init(nextY);
	mpfi_init(derivative);
	mpfi_init(integrate);

	// define and initialise left and right, will be used for plot data
	mpfr_t left, right;
	mpfr_init(left);
	mpfr_init(right);

	// assign all files
	FILE *logFile[DIMENSIONS], *pltFile[DIMENSIONS+1], *datFile[DIMENSIONS+1];

	// open logFile
	if (doLog) {
		for (i = 0; i < DIMENSIONS; i++) {
			if ((logFile[i] = fopen(logFileName[i], "w")) == NULL) {
				fprintf(stderr, "Cannot open %s\n", logFileName[i]);
				exit(0);
			}
			// writing some headers to files
			fprintf(logFile[i], "Picard: Solving %s; computing y(%2.2f).\n\n",
				equationName, MAX_TIME);
		}
	}

	// open pltFiles and datFiles
	if (doPlot) {
		for (i = 0; i < DIMENSIONS; i++) {
			if ((pltFile[i] = fopen(pltFileName[i], "w")) == NULL) {
				fprintf(stderr, "Cannot open %s\n", pltFileName[i]);
				exit(0);
			}
			if ((datFile[i] = fopen(datFileName[i], "w")) == NULL) {
				fprintf(stderr, "Cannot open %s\n", datFileName[i]);
				exit(0);
			}	
			// writing some headers to files
			fprintf(pltFile[i], "#set title \"%s\" 0,0 font \"\"\nset boxwidth 0\n", equationName);
			fprintf(datFile[i], "%-4s%-12s%-20s%-22s%s\n", "#i", "No.", "T", "X-", "X+");
		}

		// 3d plot
		if ((pltFile[DIMENSIONS] = fopen(pltFileName[DIMENSIONS], "w")) == NULL) {
			fprintf(stderr, "Cannot open %s\n", pltFileName[DIMENSIONS]);
			exit(0);
		}
		if ((datFile[DIMENSIONS] = fopen(datFileName[DIMENSIONS], "w")) == NULL) {
			fprintf(stderr, "Cannot open %s\n", datFileName[DIMENSIONS]);
			exit(0);
		}
		// writing some headers to files
		fprintf(pltFile[DIMENSIONS], "#set title \"%s\" 0,0 font \"\"\nset boxwidth 0\n", equationName);
		fprintf(datFile[DIMENSIONS], "%-4s%-12s%-20s%-22s%s\n", "#i", "No.", "T", "X", "Y");
	}

	// assign the box size for gnuplt
	double minY = -BOX_K * gnupltBoxCoeff;
	double maxY = BOX_K * gnupltBoxCoeff;
	double minX = 0;
	double maxX = MAX_TIME;
/*
	// create the first plot in pltFile
	if (doPlot) {
		for (j = 0; j < DIMENSIONS; j++) {
			fprintf(pltFile[j], "set title \"%s;\\niteration: 0\" 0,0 font \"\"\n", equationName);
			fprintf(pltFile[j], "plot [%f:%f][%f:%f] \\\n", minX, maxX, minY, maxY);
			fprintf(pltFile[j], "\t\"%s\" every :::0::0 using 3:5 title 'upper' with steps, \\\n",
				datFileName[j]);
			fprintf(pltFile[j], "\t\"%s\" every :::0::0 using 3:4 title 'lower' with steps\n",
				datFileName[j]);
			fprintf(pltFile[j], "pause -1 \"Press <ENTER> to continue to iteration 1...\"\n");
		}

		// 3d plot
		fprintf(pltFile[DIMENSIONS], "set hidden3d\n");
		fprintf(pltFile[DIMENSIONS], "set title \"%s;\\niteration: 0\" 0,0 font \"\"\n", equationName);
		//fprintf(pltFile[DIMENSIONS], "splot [%f:%f][%f:%f][%f:%f] \\\n", minX, maxX, minY, maxY, minY, maxY);
		fprintf(pltFile[DIMENSIONS], "splot '%s' every :::0::1 using 3:4:5 title 'visible' with lines\n",
			datFileName[DIMENSIONS]);
		fprintf(pltFile[DIMENSIONS], "pause -1 \"Press <ENTER> to continue to iteration 1...\"\n");
	}

	// create the first block in datFile
	if (doPlot) {
		double left_d, right_d;
		for (j = 0; j < DIMENSIONS; j++) {
			mpfi_get_left(left, currentVFvalue[j]);
			mpfi_get_right(right, currentVFvalue[j]);
			left_d = mpfr_get_d(left, mp_rnd);
			right_d = mpfr_get_d(right, mp_rnd);
			fprintf(datFile[j], "%-4d%-12d%-20.4lf%-22.4lf%-22.4lf\n", 0, 0, 0.0, left_d, right_d);
			fprintf(datFile[j], "%-4d%-12d%-20.4lf%-22.4lf%-22.4lf\n\n", 0, 1, MAX_TIME, left_d, right_d);
		}

		// 3d plot
		double left_d2, right_d2;
		fprintf(datFile[DIMENSIONS], "#next\n");

		mpfi_get_left(left, currentVFvalue[dim1]);
		mpfi_get_right(right, currentVFvalue[dim1]);
		left_d = mpfr_get_d(left, mp_rnd);
		right_d = mpfr_get_d(right, mp_rnd);
		if (right_d - left_d < MIN_RES) right_d = left_d + MIN_RES;
		mpfi_get_left(left, currentVFvalue[dim2]);
		mpfi_get_right(right, currentVFvalue[dim2]);
		left_d2 = mpfr_get_d(left, mp_rnd);
		right_d2 = mpfr_get_d(right, mp_rnd);
		if (right_d2 - left_d2 < MIN_RES) right_d2 = left_d2 + MIN_RES;
		fprintf(datFile[DIMENSIONS], "%-4d%-12d%-20.4lf%-22.4lf%-22.4lf\n", 0, 0, 0.0, right_d, left_d2);
		fprintf(datFile[DIMENSIONS], "%-4d%-12d%-20.4lf%-22.4lf%-22.4lf\n", 0, 0, 0.0, left_d, left_d2);
		fprintf(datFile[DIMENSIONS], "%-4d%-12d%-20.4lf%-22.4lf%-22.4lf\n", 0, 0, 0.0, left_d, right_d2);
		fprintf(datFile[DIMENSIONS], "%-4d%-12d%-20.4lf%-22.4lf%-22.4lf\n", 0, 0, 0.0, right_d, right_d2);
		fprintf(datFile[DIMENSIONS], "%-4d%-12d%-20.4lf%-22.4lf%-22.4lf\n\n", 0, 0, 0.0, right_d, left_d2);

		fprintf(datFile[DIMENSIONS], "%-4d%-12d%-20.4lf%-22.4lf%-22.4lf\n", 0, 1, MAX_TIME, right_d, left_d2);
		fprintf(datFile[DIMENSIONS], "%-4d%-12d%-20.4lf%-22.4lf%-22.4lf\n", 0, 1, MAX_TIME, left_d, left_d2);
		fprintf(datFile[DIMENSIONS], "%-4d%-12d%-20.4lf%-22.4lf%-22.4lf\n", 0, 1, MAX_TIME, left_d, right_d2);
		fprintf(datFile[DIMENSIONS], "%-4d%-12d%-20.4lf%-22.4lf%-22.4lf\n", 0, 1, MAX_TIME, right_d, right_d2);
		fprintf(datFile[DIMENSIONS], "%-4d%-12d%-20.4lf%-22.4lf%-22.4lf\n\n", 0, 1, MAX_TIME, right_d, left_d2);
	}
*/
	fprintf(stdout, "\nEuler: Solving %s; computing v(%2.2f, %2.2f).\n\n", equationName, MAX_TIME, MAX_TIME);

	mpfi_t M[DIMENSIONS], temp_i[DIMENSIONS];
	
	for (j = 0; j < DIMENSIONS; j++) {
		mpfi_init(temp_i[j]);
		mpfi_set(temp_i[j], box_K);
	}

	for (j = 0; j < DIMENSIONS; j++) {
		mpfi_init(M[j]);
		computeDerivativeValue(j, M[j], temp_i);
	}

	mpfr_t delta;
	mpfr_init(delta);

	// the main loop
	unsigned int n;
	for (n = 0; n <= ITERATIONS; n++) {
		fprintf(stdout, "Iteration %2d:   ", n);

		mpfr_set_ui_2exp(delta, 1, -n, mp_rnd);
		mpfr_mul(delta, delta, max_time, mp_rnd);

		for (j = 0; j < DIMENSIONS; j++) {
			mpfi_set_d(currentVFvalue[j], 0);

			// logging
			if (doLog) {
				fprintf(logFile[j], "%-3d d = \t", n);
				mpfr_out_str(logFile[j], BASE, DIGITS, delta, mp_rnd);
				fprintf(logFile[j], "\n");
				fprintf(logFile[j], "%-3d M = \t", n);
				mpfi_out_str(logFile[j], BASE, DIGITS, M[j]);
				fprintf(logFile[j], "\n");
			}

			// delta * M
			mpfi_mul_fr(temp_i[j], M[j], delta);

			// logging
			if (doLog) {
				fprintf(logFile[j], "%-3d d * M = \t", n);
				mpfi_out_str(logFile[j], BASE, DIGITS, temp_i[j]);
				fprintf(logFile[j], "\n");
			}
		}

		if (doPlot) {
			for (j = 0; j < DIMENSIONS; j++) {
				fprintf(datFile[j], "%-4d%-12d%-20.4lf%-22.4lf%-22.4lf\n", n, 0, 0.0, 0.0, 0.0);
			}
		}

		unsigned long precision = 1 << n; // 2^n
		//fprintf(stdout, "precision = %ld ", precision);

		for (k = 1; k <= precision; k++) {
			/*
			fprintf(logFile[0], "q[%d] = ", k);
			mpfr_mul_ui(q, delta, k, mp_rnd);
			mpfr_out_str(logFile[0], BASE, DIGITS, q, mp_rnd);
			fprintf(logFile[0], "\n");
			*/

			for (j = 0; j < DIMENSIONS; j++) {
				// logging
				if (doLog) {
					fprintf(logFile[j], "%-3d %-10d y = \t", n, k);
					mpfi_out_str(logFile[j], BASE, DIGITS, currentVFvalue[j]);
					fprintf(logFile[j], "\n");
				}

				// curr + delta * M
				mpfi_add(temp_i[j], temp_i[j], currentVFvalue[j]);
	
				// logging
				if (doLog) {
					fprintf(logFile[j], "%-3d %-10d y + d * M = \t", n, k);
					mpfi_out_str(logFile[j], BASE, DIGITS, temp_i[j]);
					fprintf(logFile[j], "\n");
				}

				// next = v (curr + delta * M)
				computeDerivativeValue(j, nextVFvalue[j], temp_i);

				// logging
				if (doLog) {
					fprintf(logFile[j], "%-3d %-10d v (y + d * M) = \t", n, k);
					mpfi_out_str(logFile[j], BASE, DIGITS, nextVFvalue[j]);
					fprintf(logFile[j], "\n");
				}

				// next = delta * v (curr + delta * M)
				mpfi_mul_fr(nextVFvalue[j], nextVFvalue[j], delta);

				// logging
				if (doLog) {
					fprintf(logFile[j], "%-3d %-10d d * v (y + d * M) = \t", n, k);
					mpfi_out_str(logFile[j], BASE, DIGITS, nextVFvalue[j]);
					fprintf(logFile[j], "\n");
				}

				// next = curr + delta * v (curr + delta * M)
				mpfi_add(nextVFvalue[j], nextVFvalue[j], currentVFvalue[j]);

				// intersect nextY with [-BOX_K, BOX_K]
				int res = mpfi_intersect(nextVFvalue[j], nextVFvalue[j], box_K);
				if (res) printf ("%d", res);

				// logging
				if (doLog) {
					fprintf(logFile[j], "%-3d %-10d y + d * v (y + d * M) = \t", n, k);
					mpfi_out_str(logFile[j], BASE, DIGITS, nextVFvalue[j]);
					fprintf(logFile[j], "\n");
				}
			}
			if (doLog) for (j = 0; j < DIMENSIONS; j++) fprintf(logFile[j], "\n");

			// create a block in datFile
			if (doPlot) {
				double left_d, right_d;
				for (j = 0; j < DIMENSIONS; j++) {
					mpfi_get_left(left, nextVFvalue[j]);
					mpfi_get_right(right, nextVFvalue[j]);
					left_d = mpfr_get_d(left, mp_rnd);
					right_d = mpfr_get_d(right, mp_rnd);
					fprintf(datFile[j], "%-4d%-12d%-20.4lf%-22.4lf%-22.4lf\n",
						n, k, MAX_TIME * k / precision, left_d, right_d);
				}
/*
				// 3d plot
				double left_d2, right_d2;
				fprintf(datFile[DIMENSIONS], "#next\n");
				for (k = 0; k < precision; k++) {
					mpfi_get_left(left, nextVFvalue[k][dim1]);
					mpfi_get_right(right, nextVFvalue[k][dim1]);
					left_d = mpfr_get_d(left, mp_rnd);
					right_d = mpfr_get_d(right, mp_rnd);
					if (right_d - left_d < MIN_RES) right_d = left_d + MIN_RES;
					mpfi_get_left(left, nextVFvalue[k][dim2]);
					mpfi_get_right(right, nextVFvalue[k][dim2]);
					left_d2 = mpfr_get_d(left, mp_rnd);
					right_d2 = mpfr_get_d(right, mp_rnd);
					if (right_d2 - left_d2 < MIN_RES) right_d2 = left_d2 + MIN_RES;
					fprintf(datFile[DIMENSIONS], "%-4d%-12d%-20.4lf%-22.4lf%-22.4lf\n",
						n, k, MAX_TIME * k / precision, right_d, left_d2);
					fprintf(datFile[DIMENSIONS], "%-4d%-12d%-20.4lf%-22.4lf%-22.4lf\n",
						n, k, MAX_TIME * k / precision, left_d, left_d2);
					fprintf(datFile[DIMENSIONS], "%-4d%-12d%-20.4lf%-22.4lf%-22.4lf\n",
						n, k, MAX_TIME * k / precision, left_d, right_d2);
					fprintf(datFile[DIMENSIONS], "%-4d%-12d%-20.4lf%-22.4lf%-22.4lf\n",
						n, k, MAX_TIME * k / precision, right_d, right_d2);
					fprintf(datFile[DIMENSIONS], "%-4d%-12d%-20.4lf%-22.4lf%-22.4lf\n\n",
						n, k, MAX_TIME * k / precision, right_d, left_d2);
				}
				fprintf(datFile[j], "%-4d%-12d%-20.4lf%-22.4lf%-22.4lf\n",
					n, precision, MAX_TIME, right_d, left_d2);
				fprintf(datFile[j], "%-4d%-12d%-20.4lf%-22.4lf%-22.4lf\n",
					n, precision, MAX_TIME, left_d, left_d2);
				fprintf(datFile[j], "%-4d%-12d%-20.4lf%-22.4lf%-22.4lf\n",
					n, precision, MAX_TIME, left_d, right_d2);
				fprintf(datFile[j], "%-4d%-12d%-20.4lf%-22.4lf%-22.4lf\n",
					n, precision, MAX_TIME, right_d, right_d2);
				fprintf(datFile[j], "%-4d%-12d%-20.4lf%-22.4lf%-22.4lf\n\n",
					n, precision, MAX_TIME, right_d, left_d2);
*/
			}

			for (j = 0; j < DIMENSIONS; j++) mpfi_set(currentVFvalue[j], nextVFvalue[j]);

		} // for k

		for (j = 0; j < DIMENSIONS; j++) {
			// report precision on stdout		
			fprintf(stdout, "V%d: ", j);
			mpfi_out_str(stdout, BASE, DIGITS, currentVFvalue[j]);
			fprintf(stdout, ", Precision: ");
			mpfi_diam_abs(temp_r, currentVFvalue[j]);
			mpfr_out_str(stdout, BASE, DIGITS, temp_r, mp_rnd);
			fprintf(stdout, "\n\t\t");
		}
		fprintf(stdout, "\n");

		// create a plot in pltFile
		if (doPlot) {
			for (j = 0; j < DIMENSIONS; j++) {
				mpfi_get_left(left, currentVFvalue[j]);
				mpfi_get_right(right, currentVFvalue[j]);
				minY = mpfr_get_d(left, mp_rnd);
				maxY = mpfr_get_d(right, mp_rnd);
				minY *= gnupltBoxCoeff;
				maxY *= gnupltBoxCoeff;
				fprintf(pltFile[j], "set title \"%s;\\niteration: %d\" 0,0 font \"\"\n", equationName, n);
				//fprintf(pltFile[j], "plot [%f:%f][%f:%f] \\\n", minX, maxX, minY, maxY);
				fprintf(pltFile[j], "plot \\\n");
				fprintf(pltFile[j], "\t\"%s\" every :::%d::%d using 3:5 title 'upper' with lines, \\\n",
					datFileName[j], n, n);
				fprintf(pltFile[j], "\t\"%s\" every :::%d::%d using 3:4 title 'lower' with lines, \\\n",
					datFileName[j], n, n);
				fprintf(pltFile[j], "\t\"%s\" every :::%d::%d using 3:4:4:5:5 title 'value' with candlesticks\n",
					datFileName[j], n, n);
				if (n < ITERATIONS) fprintf(pltFile[j],
					"pause -1 \"Press <ENTER> to continue to iteration %d...\"\n", n+1);
				fprintf(datFile[j], "\n"); //separate the blocks
			}
/*
			// 3d plot
			fprintf(pltFile[DIMENSIONS], "set title \"%s;\\niteration: %d\" 0,0 font \"\"\n", equationName, n);
			//fprintf(pltFile[DIMENSIONS], "splot [%f:%f][%f:%f][%f:%f] \\\n", minX, maxX, minY, maxY, minY, maxY);
			fprintf(pltFile[DIMENSIONS], "splot \"%s\" every :%d::%d::%d using 3:4:5 title 'visible' with lines\n",
				datFileName[DIMENSIONS], (n > 8 ? (1 << (n - 8)) : 1), (1 << n) + n - 1, (1 << n) + n - 1 + (1 << n));
			if (n < ITERATIONS) fprintf(pltFile[DIMENSIONS],
				"pause -1 \"Press <ENTER> to continue to iteration %d...\"\n", n+1);
*/
		} // doPlot

		if (doLog) for (j = 0; j < DIMENSIONS; j++) fprintf(logFile[j], "\n");
/*
		for (j = 0; j < DIMENSIONS; j++) {
			mpfi_set_d(currentY, 0);
			mpfi_set_d(nextY, 0);

			for (k = 0; k < precision; k++) {
				kk = k >> 1; // k / 2

				// logging
				if (doLog) {
					fprintf(logFile[j], "iteration: %d, partition: %d\n", n, k);
					fprintf(logFile[j], "currentVFvalue%d[%d]: ", j, kk);
					mpfi_out_str(logFile[j], BASE, DIGITS, currentVFvalue[kk][j]);
					fprintf(logFile[j], "\n");
				}

				// compute derivative for the current partition
				computeDerivativeValue(j, derivative, currentVFvalue[kk]);

				// logging
				if (doLog) {
					fprintf(logFile[j], "\tderivative%d: ", j);
					mpfi_out_str(logFile[j], BASE, DIGITS, derivative);
					fprintf(logFile[j], "\n");
				}

				// integrate; note that the width of the partition is 2^(-n) * MAX_TIME
				mpfi_mul_2si(integrate, derivative, -n);
				mpfi_mul_d(integrate, integrate, MAX_TIME);

				// logging
				if (doLog) {
					fprintf(logFile[j], "\tintegrate%d: ", j);
					mpfi_out_str(logFile[j], BASE, DIGITS, integrate);
					fprintf(logFile[j], "\n");
				}

				// add the result of integrate to the currentY positiuon
				mpfi_add(nextY, integrate, currentY);

				// logging
				if (doLog) {
					fprintf(logFile[j], "\tcurrentY%d: ", j);
					mpfi_out_str(logFile[j], BASE, DIGITS, currentY);
					fprintf(logFile[j], "\n");

					fprintf(logFile[j], "\tnextY%d: ", j);
					mpfi_out_str(logFile[j], BASE, DIGITS, nextY);
					fprintf(logFile[j], "\n");
				}

				// intersect nextY with [-BOX_K, BOX_K]
				int res = mpfi_intersect(nextY, nextY, box_K);
				if (res) printf ("%d", res);

				// logging
				if (doLog) {
					fprintf(logFile[j], "\tTrunc nextY%d: ", j);
					mpfi_out_str(logFile[j], BASE, DIGITS, nextY);
					fprintf(logFile[j], "\n");
				}

				// store the convex hull of currentY and nextY into nextVFvalue[k]
				mpfi_union(nextVFvalue[k][j], currentY, nextY);

				// logging
				if (doLog) {
					fprintf(logFile[j], "\tnextVFvalue%d[%d]: ", j, k);
					mpfi_out_str(logFile[j], BASE, DIGITS, nextVFvalue[k][j]);
					fprintf(logFile[j], "\n\n");
				}

				// update currentY with nextY
				mpfi_set(currentY, nextY);
			}

			// report precision on stdout		
			fprintf(stdout, "V%d: ", j);
			mpfi_out_str(stdout, BASE, DIGITS, currentY);
			fprintf(stdout, ", Precision: ");
			mpfi_diam_abs(temp_r, currentY);
			mpfr_out_str(stdout, BASE, DIGITS, temp_r, mp_rnd);
			fprintf(stdout, "\n\t\t");
		}
		fprintf(stdout, "\n");

		// create a plot in pltFile
		if (doPlot) {
			for (j = 0; j < DIMENSIONS; j++) {
				mpfi_get_left(left, nextVFvalue[0][j]);
				mpfi_get_right(right, nextVFvalue[0][j]);
				minY = mpfr_get_d(left, mp_rnd);
				maxY = mpfr_get_d(right, mp_rnd);
				double min, max;
				for (k = 1; k < precision; k++) {
					mpfi_get_left(left, nextVFvalue[k][j]);
					mpfi_get_right(right, nextVFvalue[k][j]);
					min = mpfr_get_d(left, mp_rnd);
					if (min < minY) minY = min;
					max = mpfr_get_d(right, mp_rnd);
					if (max > maxY) maxY = max;
				}
				minY *= gnupltBoxCoeff;
				maxY *= gnupltBoxCoeff;
				fprintf(pltFile[j], "set title \"%s;\\niteration: %d\" 0,0 font \"\"\n", equationName, n);
				fprintf(pltFile[j], "plot [%f:%f][%f:%f] \\\n", minX, maxX, minY, maxY);
				fprintf(pltFile[j], "\t\"%s\" every :::%d::%d using 3:5 title 'upper' with steps, \\\n",
					datFileName[j], n, n);
				fprintf(pltFile[j], "\t\"%s\" every :::%d::%d using 3:4 title 'lower' with steps\n",
					datFileName[j], n, n);
				if (n < ITERATIONS) fprintf(pltFile[j],
					"pause -1 \"Press <ENTER> to continue to iteration %d...\"\n", n+1);
			}

			// 3d plot
			fprintf(pltFile[DIMENSIONS], "set title \"%s;\\niteration: %d\" 0,0 font \"\"\n", equationName, n);
			//fprintf(pltFile[DIMENSIONS], "splot [%f:%f][%f:%f][%f:%f] \\\n", minX, maxX, minY, maxY, minY, maxY);
			fprintf(pltFile[DIMENSIONS], "splot \"%s\" every :%d::%d::%d using 3:4:5 title 'visible' with lines\n",
				datFileName[DIMENSIONS], (n > 8 ? (1 << (n - 8)) : 1), (1 << n) + n - 1, (1 << n) + n - 1 + (1 << n));
			if (n < ITERATIONS) fprintf(pltFile[DIMENSIONS],
				"pause -1 \"Press <ENTER> to continue to iteration %d...\"\n", n+1);
		}

		// create a block in datFile
		if (doPlot) {
			double left_d, right_d;
			for (j = 0; j < DIMENSIONS; j++) {
				for (k = 0; k < precision; k++) {
					mpfi_get_left(left, nextVFvalue[k][j]);
					mpfi_get_right(right, nextVFvalue[k][j]);
					left_d = mpfr_get_d(left, mp_rnd);
					right_d = mpfr_get_d(right, mp_rnd);
					fprintf(datFile[j], "%-4d%-12d%-20.4lf%-22.4lf%-22.4lf\n",
						n, k, MAX_TIME * k / precision, left_d, right_d);
				}
				fprintf(datFile[j], "%-4d%-12d%-20.4lf%-22.4lf%-22.4lf\n\n",
					n, precision, MAX_TIME, left_d, right_d);
			}

			// 3d plot
			double left_d2, right_d2;
			fprintf(datFile[DIMENSIONS], "#next\n");
			for (k = 0; k < precision; k++) {
				mpfi_get_left(left, nextVFvalue[k][dim1]);
				mpfi_get_right(right, nextVFvalue[k][dim1]);
				left_d = mpfr_get_d(left, mp_rnd);
				right_d = mpfr_get_d(right, mp_rnd);
				if (right_d - left_d < MIN_RES) right_d = left_d + MIN_RES;
				mpfi_get_left(left, nextVFvalue[k][dim2]);
				mpfi_get_right(right, nextVFvalue[k][dim2]);
				left_d2 = mpfr_get_d(left, mp_rnd);
				right_d2 = mpfr_get_d(right, mp_rnd);
				if (right_d2 - left_d2 < MIN_RES) right_d2 = left_d2 + MIN_RES;
				fprintf(datFile[DIMENSIONS], "%-4d%-12d%-20.4lf%-22.4lf%-22.4lf\n",
					n, k, MAX_TIME * k / precision, right_d, left_d2);
				fprintf(datFile[DIMENSIONS], "%-4d%-12d%-20.4lf%-22.4lf%-22.4lf\n",
					n, k, MAX_TIME * k / precision, left_d, left_d2);
				fprintf(datFile[DIMENSIONS], "%-4d%-12d%-20.4lf%-22.4lf%-22.4lf\n",
					n, k, MAX_TIME * k / precision, left_d, right_d2);
				fprintf(datFile[DIMENSIONS], "%-4d%-12d%-20.4lf%-22.4lf%-22.4lf\n",
					n, k, MAX_TIME * k / precision, right_d, right_d2);
				fprintf(datFile[DIMENSIONS], "%-4d%-12d%-20.4lf%-22.4lf%-22.4lf\n\n",
					n, k, MAX_TIME * k / precision, right_d, left_d2);
			}
			fprintf(datFile[j], "%-4d%-12d%-20.4lf%-22.4lf%-22.4lf\n",
				n, precision, MAX_TIME, right_d, left_d2);
			fprintf(datFile[j], "%-4d%-12d%-20.4lf%-22.4lf%-22.4lf\n",
				n, precision, MAX_TIME, left_d, left_d2);
			fprintf(datFile[j], "%-4d%-12d%-20.4lf%-22.4lf%-22.4lf\n",
				n, precision, MAX_TIME, left_d, right_d2);
			fprintf(datFile[j], "%-4d%-12d%-20.4lf%-22.4lf%-22.4lf\n",
				n, precision, MAX_TIME, right_d, right_d2);
			fprintf(datFile[j], "%-4d%-12d%-20.4lf%-22.4lf%-22.4lf\n\n",
				n, precision, MAX_TIME, right_d, left_d2);
		}

		// set currentVFvalue to be nextVFvalue
		for (j = 0; j < DIMENSIONS; j++) {
			for (k = 0; k < precision; k++) {
				mpfi_set(currentVFvalue[k][j], nextVFvalue[k][j]);
			}
		}*/
	}

	// logging
	if (doPlot) {
		for (j = 0; j <= DIMENSIONS; j++) {
			fprintf(pltFile[j], "pause -1 \"There is no more refinement. Press <ENTER> to exit.\"\n");
		}
	}

	//
	// clearing up
	//

	for (j = 0; j < DIMENSIONS; j++) {
		// close all files
		if (doLog) {
			fclose(logFile[j]);
		}
		if (doPlot) {
			fclose(pltFile[j]);
			fclose(datFile[j]);
		}

		// clear all mpfi_t's
		mpfi_clear(currentVFvalue[j]);
		mpfi_clear(nextVFvalue[j]);
		mpfi_clear(M[j]);
		mpfi_clear(temp_i[j]);
	}
	if (doPlot) {
		fclose(pltFile[DIMENSIONS]);
		fclose(datFile[DIMENSIONS]);
	}

	mpfi_clear(currentY);
	mpfi_clear(nextY);
	mpfi_clear(derivative);
	mpfi_clear(integrate);
	mpfi_clear(box_K);

	// clear all mpfr_t's
	mpfr_clear(left);
	mpfr_clear(right);
	mpfr_clear(temp_r);
	mpfr_clear(delta);
	mpfr_clear(max_time);

	// free memory used for file names
	for (j = 0; j < DIMENSIONS; j++) {
		free(logFileName[j]);
		free(datFileName[j]);
		free(pltFileName[j]);
	}
	free(logFileName);
	free(datFileName);
	free(pltFileName);

	return 0;
}
